"""
Copyright (C) 2016, 2017, 2018 Cumulus Networks, inc.

All Rights reserved.

This software is subject to the Cumulus Networks End User License Agreement available
at the following locations:.

Internet: https://cumulusnetworks.com/downloads/eula/latest/view/

Cumulus Linux systems: /usr/share/cumulus/EULA.txt


This module is permissions sensitive.  Its configuration files contain passwords that
should not be shown to users without edit permission.
"""

from fnmatch import fnmatch
from nclu import (
    WORKING_DIRECTORIES_PERMISSIONS,
    chown_chgroup,
    files_match,
    get_pending_add_del_commands,
    glob_to_numbers,
    ifname_expand_glob,
    service_is_active,
    service_is_enabled,
)
from nclu.plugins import (
    get_config_files_helper,
    get_pending_diffs,
    merge_messages,
    persist_configurations,
    remove_scratch_files_and_backups,
)
from nclu.plugins.frr.adddel import execute_command_add_del
from nclu.plugins.frr.core import (
    ETC_FRR, ETC_FRR_CONF,
    ETC_FRR_CONF_BACKUP,
    ETC_FRR_CONF_BASELINE,
    ETC_FRR_CONF_DEFAULT,
    ETC_FRR_CONF_SCRATCHPAD,
    ETC_FRR_DAEMONS,
    ETC_FRR_DAEMONS_BACKUP,
    ETC_FRR_DAEMONS_BASELINE,
    ETC_FRR_DAEMONS_SCRATCHPAD,
    FRR_RELOAD_PY,
    OWNER_NAME,
    SCRATCHPAD_DIRECTORY,
    SERVICE_NAME,
    get_daemons,
    frr_wr_mem,
    frr_restart_needed,
    frr_restart,
    init_guts,
    init_guts_add_del,
    file_baselines_match,
)
from nclu.plugins.frr.frr_docstring import __doc__  # noqa
from nclu.plugins.frr.showclear import (
    execute_command_show_clear,
    sanitize_passwords,
    show_config_commands as showclear_show_config_commands,
    show_config_summary as showclear_show_config_summary,
)
from nclu.utils import check_output
from subprocess import STDOUT, CalledProcessError
import doctest
import logging
import os
import re
import shutil


# NetDaemon.reload looks for this.
per_command_helptext = {
    "net (add|del) bgp timers": {"<0-65535>": "Keepalive interval"},
    "net (add|del) bgp timers <0-65535>": {"<0-65535>": "Holdtime"},
    "net (add|del) bgp neighbor <bgppeer> bfd": {"<2-255>": "Detect Multiplier"},
    "net (add|del) bgp neighbor <bgppeer> bfd <2-255>": {"<50-60000>": "Required min receive interval"},
    "net (add|del) bgp neighbor <bgppeer> bfd <2-255> <50-60000>": {"<50-60000>": "Desired min transmit interval"},
}

logging.getLogger("frr-reload").setLevel(logging.WARNING)
log = logging.getLogger("netd")


def frr_reload(verbose):
    messages = {}

    try:
        output = check_output([FRR_RELOAD_PY, "--overwrite", "--stdout", "--reload", ETC_FRR_CONF], stderr=STDOUT)
        if verbose:
            messages["output"] = output

        result, wr_mem_messages = frr_wr_mem()
        merge_messages(messages, wr_mem_messages)

    except CalledProcessError as e:
        log.exception(e)
        log.error(e.output)
        messages["error"] = ["{0} failed.  See /var/log/netd.log for details.".format(FRR_RELOAD_PY)]
        return (False, messages)

    return (result, messages)


def copy_scratchpad_to_frr_conf():
    """
    FRR is picky about frr.conf being "marked".  Remove all of the "end"
    lines that are there as a result of marking the contexts and sub-contexts.
    """

    with open(ETC_FRR_CONF, 'w') as fh_new, open(ETC_FRR_CONF_SCRATCHPAD) as fh_scratch:
        for line in fh_scratch:
            if line.strip() != "end":
                fh_new.write(line)

    chown_chgroup(ETC_FRR_CONF, OWNER_NAME, OWNER_NAME)


def commit_pending(verbose, net_del_all):
    """
    Copy scratchpad files over configuration files, call necessary restart commands, etc.

    Return a 2-tuple of (bool, dict).  The boolean indicates if the commit succeeded.
    The dictionary is as described in nclu/plugins/__init__.py:execute_modular.

    net_del_all - A boolean; if True, the user is committing "net del all"; this plug-in doesn't use it
    """
    if not os.path.isfile(ETC_FRR_DAEMONS_SCRATCHPAD) and not os.path.isfile(ETC_FRR_CONF_SCRATCHPAD):
        # There are no changes to commit.
        return (True, {})

    if (files_match(ETC_FRR_DAEMONS_BASELINE, ETC_FRR_DAEMONS_SCRATCHPAD) and
        files_match(ETC_FRR_CONF_BASELINE, ETC_FRR_CONF_SCRATCHPAD)):

        # There are no changes to commit.
        remove_scratch_files_and_backups((ETC_FRR_DAEMONS_SCRATCHPAD, ETC_FRR_CONF_SCRATCHPAD))
        return (True, {})

    # Did user-A edit the config file after user-B started staging changes via nclu?
    result, messages = file_baselines_match()
    if not result:
        return (False, messages)

    # A service must be enabled to start automatically following a reboot.
    if not service_is_enabled(SERVICE_NAME):
        try:
            check_output(["/bin/systemctl", "enable", SERVICE_NAME])
        except CalledProcessError as e:
            log.exception(e)
            log.error(e.output)
            assert "error" not in messages, "If there is already an error message, the code should have returned above, after the call to file_baselines_match."
            messages["error"] = ["Failed to enable {0}.  See /var/log/netd.log for more details.".format(SERVICE_NAME)]
            return (False, messages)

    # This function call must preceed copying over ETC_FRR_DAEMONS, which happens below in a call to persist_configurations.
    do_frr_restart = frr_restart_needed()

    if os.path.isfile(ETC_FRR_DAEMONS_SCRATCHPAD):
        # Copy the daemons scratchpad file over the real daemons file.

        persist_configurations([(ETC_FRR_DAEMONS, ETC_FRR_DAEMONS_SCRATCHPAD)])
        chown_chgroup(ETC_FRR_DAEMONS, OWNER_NAME, OWNER_NAME)

    if os.path.isfile(ETC_FRR_CONF_SCRATCHPAD):
        # Copy the frr.conf scratchpad file over the real frr.conf file.  Make a backup, if possible.

        if os.path.isfile(ETC_FRR_CONF):
            shutil.copyfile(ETC_FRR_CONF, ETC_FRR_CONF_BACKUP)
        copy_scratchpad_to_frr_conf()

    # TODO - we need the diff (Why?)
    if do_frr_restart:
        result, restart_messages = frr_restart(verbose)
        merge_messages(messages, restart_messages)

        if result:
            result, frr_wr_mem_messages = frr_wr_mem()
            merge_messages(messages, frr_wr_mem_messages)

            if result:
                copy_scratchpad_to_frr_conf()
                result, reload_messages = frr_reload(verbose)
                merge_messages(messages, reload_messages)
    else:
        result, reload_messages = frr_reload(verbose)
        merge_messages(messages, reload_messages)

    if result:
        # Before deleting the scratchpad, copy FRR configuration to the default file so we can use it next time.
        if os.path.isfile(ETC_FRR_CONF_SCRATCHPAD) and os.path.isfile(ETC_FRR_CONF):
            with open(ETC_FRR_CONF_SCRATCHPAD) as f:
                new_frr_conf = f.read().strip()
            if new_frr_conf == _nclu_default_config.strip():
                # We were on a del all command commit.  Copy over the new FRR configuration.
                with open(ETC_FRR_CONF) as frr_f:
                    with open(ETC_FRR_CONF_DEFAULT, 'w') as default_f:
                        default_f.write(frr_f.read())

        # reload/restart worked.  Delete backup and scratchpad files.
        remove_scratch_files_and_backups((ETC_FRR_DAEMONS_SCRATCHPAD, ETC_FRR_CONF_SCRATCHPAD))
    else:
        # reload/restart failed.  Restore the configuration files from the backups, but do
        # not delete the scratchpad files.

        if os.path.isfile(ETC_FRR_DAEMONS_BACKUP):
            persist_configurations([(ETC_FRR_DAEMONS, ETC_FRR_DAEMONS_BACKUP)], False)
            os.remove(ETC_FRR_DAEMONS_BACKUP)

        if os.path.isfile(ETC_FRR_CONF_BACKUP):
            persist_configurations([(ETC_FRR_CONF, ETC_FRR_CONF_BACKUP)], False)
            os.remove(ETC_FRR_CONF_BACKUP)

        # restart/reload frr to apply the original configuration.
        # Don't alter the value of "result".  The commit has already failed.
        if do_frr_restart:
            _, restart_messages = frr_restart(verbose)
            merge_messages(messages, restart_messages)
        else:
            _, reload_messages = frr_reload(verbose)
            merge_messages(messages, reload_messages)

    return (result, messages)


def execute_command(net_daemon):
    """
    Given an instance of NetDaemon that has already run an input command through NetworkDocopt,
    perform the appropriate actions, and return the standard 2-tuple.
    """

    # Convenience variable
    tokens = net_daemon.nclu_parser.args
    assert tokens.get("configuration") is None, "\"net show configuration\" does not use this code path."

    result, messages = init_guts()

    if not result:
        return (False, messages)

    ifnames = []
    bgppeers = []
    result = True

    if tokens.get("<interface>"):
        ifname = str(tokens.get("<interface>"))

        try:
            ifnames = ifname_expand_glob(ifname)  # A list
        except ValueError as e:
            messages["error"] = ["\"{0}\" is not a valid interface name nor glob.  {1}".format(ifname, e)]
            log.error(messages["error"][0])
            return (False, messages)

    elif tokens.get("vlan") and tokens.get("<number-range-list>"):
        vlan_numbers = glob_to_numbers(str(tokens.get("<number-range-list>")))

        for number in vlan_numbers:
            ifnames.append("vlan{0}".format(number))

    elif tokens.get("<bgppeer>"):
        bgppeer = str(tokens.get("<bgppeer>"))

        try:
            bgppeers.extend(ifname_expand_glob(bgppeer))
        except ValueError as e:
            messages["error"] = ["\"{0}\" is not a valid interface name nor glob.  {1}".format(bgppeer, e)]
            log.error(messages["error"][0])
            return (False, messages)

    if tokens.get('add') or tokens.get('del'):
        (result, messages) = init_guts_add_del()

        if result:
            (result, messages) = execute_command_add_del(net_daemon, ifnames, bgppeers)

    elif tokens.get("show") or tokens.get("clear"):
        (result, messages) = execute_command_show_clear(net_daemon, ifnames, bgppeers)

        if "is not running" in messages.get("output", ''):
            # A service managed by frr.service, such as pimd or ospfd, is not running.
            result = False
            handle_service_notifications(messages)

    else:
        messages["error"] = ["No commands matched."]
        result = False

    return (result, messages)


def handle_service_notifications(messages):
    """
    Modify a messages dictionary containing notice of services not running by removing
    the notifications from "output", deduplicating, and moving them to "errors".
    """

    assert "output" in messages
    assert "is not running" in messages["output"]

    if "error" not in messages:
        messages["error"] = []

    services_not_running = frozenset(re.findall(r"""^\s*\S+ is not running$""", messages["output"], re.MULTILINE))

    for notification in services_not_running:
        messages["output"] = ''
        messages["error"].append(notification)

    messages["output"] = messages["output"].strip()


# predicted frr generated default config that can be checked
# against fnmatch to guess if current config is default generated
# from frr. This should only happen if we don't have a stored default
_frr_default_config_lines = (
    'frr version *',
    'frr defaults datacenter',
    'hostname *',
    'username *',
    'password *',
    'service integrated-vtysh-config',
    'line vty',
    'exec-timeout 0 0',
    '!')

# nclu provided default config
# NOTE: we can not provide a full default config
# because parts of it are dynamically modified by frr
# after config is applied.
_nclu_default_config = "service integrated-vtysh-config\n"
_nclu_default_daemons = "zebra=yes\n"


def _is_default_config():
    """
    If we do not have a default config already saved, this
    provides a "smart" way to check if the current config
    is already the default frr config
    """
    # first, check for reverted file type.
    # frr writes special configs when you try to reset it
    # that are dynamic for each system
    if not os.path.isfile(ETC_FRR_CONF):
        return False
    with open(ETC_FRR_CONF) as fi:
        for line in fi:
            line = line.strip()
            found = False
            for default_line in _frr_default_config_lines:
                if fnmatch(line, default_line):
                    found = True
                    break
            if not found:
                return False
        else:
            # went through every line verifying this is the default config
            return True
    return True


def del_all():
    """
    net del all
    """

    if not os.path.isdir(SCRATCHPAD_DIRECTORY):
        os.makedirs(SCRATCHPAD_DIRECTORY, mode=WORKING_DIRECTORIES_PERMISSIONS)

    if os.path.isfile(ETC_FRR_DAEMONS):
        shutil.copyfile(ETC_FRR_DAEMONS, ETC_FRR_DAEMONS_BASELINE)
    else:
        with open(ETC_FRR_DAEMONS, 'w') as fi:
            fi.write(_nclu_default_daemons)
        with open(ETC_FRR_DAEMONS_BASELINE, 'w') as fi:
            fi.write(_nclu_default_daemons)

    with open(ETC_FRR_DAEMONS_SCRATCHPAD, 'w') as f:
        f.write(_nclu_default_daemons)

    if not os.path.isfile(ETC_FRR_CONF_DEFAULT) and _is_default_config():
        return

    if os.path.isfile(ETC_FRR_CONF):
        shutil.copyfile(ETC_FRR_CONF, ETC_FRR_CONF_BASELINE)
    else:
        with open(ETC_FRR_CONF, 'w') as fi:
            fi.write(_nclu_default_config)
        with open(ETC_FRR_CONF_BASELINE, 'w') as fi:
            fi.write(_nclu_default_config)

    if os.path.isfile(ETC_FRR_CONF_DEFAULT):
        shutil.copyfile(ETC_FRR_CONF_DEFAULT, ETC_FRR_CONF_SCRATCHPAD)
        # The generated default config will not be complete.  We'll work
        # around that here by tacking on things that need to be explicitly
        # set back to the defaults.  Longer term, we should either
        #  A) Ask FRR to generate it's complete default (if that's possible).
        #  B) Skip the "wr mem" after resetting to defaults.  This would
        #     require an FRR restart though, but that's probably OK given it's
        #     a del all and implies a disruptive operation.
        with open(ETC_FRR_CONF_SCRATCHPAD, 'a') as fi:
            fi.write("no exec-timeout\n")
    else:
        with open(ETC_FRR_CONF_SCRATCHPAD, 'w') as f:
            # fallback to our unknown default config
            f.write(_nclu_default_config)


def get_config_files(user_may_edit):
    """
    Return a string with the content of all existing, non-empty configuration files managed with this plug-in.
    """

    managed_files_text = get_config_files_helper(get_managed_files())

    if user_may_edit:
        return managed_files_text

    # The user doesn't have edit permission.  Redact sensitive information.
    return sanitize_passwords(managed_files_text)


def get_pending(use_colors):
    return get_pending_diffs([(ETC_FRR_DAEMONS, ETC_FRR_DAEMONS_SCRATCHPAD),
                              (ETC_FRR_CONF_BASELINE, ETC_FRR_CONF_SCRATCHPAD)],
                             use_colors)


def abort_pending():
    remove_scratch_files_and_backups((ETC_FRR_DAEMONS_SCRATCHPAD, ETC_FRR_CONF_SCRATCHPAD))


def get_managed_files():
    return [ETC_FRR_DAEMONS, ETC_FRR_CONF]


def show_config_commands(user_may_edit, ifupdown2_wrapper):
    return showclear_show_config_commands(user_may_edit, ifupdown2_wrapper)


def show_config_summary(user_may_edit, summary):
    showclear_show_config_summary(user_may_edit, summary)


def get_restart_commands(filename, diff_lines):
    """
    This function is only called during a rollback; commit_pending doesn't use it.
    """

    assert filename in get_managed_files(), "filename = {0}".format(filename)
    assert isinstance(diff_lines, list) and diff_lines, "This should be a non-empty list."

    if filename == ETC_FRR_DAEMONS:
        action = frr_restart_needed_using_diff_lines(diff_lines)

        if action is None:
            return []

        if action == "RESTART":
            return [("/bin/systemctl", "enable", SERVICE_NAME),
                    ("/bin/systemctl", "reset-failed", SERVICE_NAME),
                    ("/bin/systemctl", "restart", SERVICE_NAME)]

        if action == "STOP":
            return [("/bin/systemctl", "disable", SERVICE_NAME),
                    ("/bin/systemctl", "reset-failed", SERVICE_NAME),
                    ("/bin/systemctl", "stop", SERVICE_NAME)]

        # This is a programming error.
        raise RuntimeError("frr_restart_needed_using_diff_lines returned {0}.".format(action))

    assert filename == ETC_FRR_CONF, "This is the only other option right now."

    if diff_lines:
        # A reload is required (unless a "restart" or "stop" is to occur; this is determined upstream, in rollback.py).
        return [("/bin/systemctl", "enable", SERVICE_NAME),
                ("/bin/systemctl", "reset-failed", SERVICE_NAME),
                ("/bin/systemctl", "reload", SERVICE_NAME)]

    return []


def frr_restart_needed_using_diff_lines(diff_lines):
    """
    Given the diff for ETC_FRR_DAEMONS as computed in rollback.py:rollback, return "RESTART" if FRR should be restarted;
    "STOP", if FRR should be stopped; or None, if taking no action is appropriate.

    The criteria for returning "RESTART" is that any service is being enabled (set to "yes").  If no service is being
    enabled, and some will remain running, return None.

    Return "STOP" if the user is stopping "zebra".

    This function should never be called if ETC_FRR_DAEMONS is not changing.

    diff_lines - A list of strings, like:

        --- /etc/frr/daemons	2018-09-13 21:30:20.000000000 +0000
        +++ /run/nclu/frr/daemons.scratchpad	2018-09-18 17:48:20.096068162 +0000
        @@ -1,37 +1 @@
        ...
        -zebra=no
        -bgpd=no
        ...
        +zebra=yes
    """

    # Enabling a service
    yes_re = re.compile(r"""^\s*\+\s*\S+\s*=\s*yes\b""", re.IGNORECASE | re.MULTILINE)

    # Disabling FRR.  Stopping a service is indicated by removing "<service>=yes".
    no_frr_re = re.compile(r"""^\s*\-\s*frr\s*=\s*yes\b""", re.IGNORECASE | re.MULTILINE)

    for line in diff_lines:
        if yes_re.match(line) is not None:
            # A service is being enabled.
            return "RESTART"

        if no_frr_re.match(line) is not None:
            # FRR is being disabled.
            return "STOP"

    # Determine if any services are still running.  Note that ETC_FRR_DAEMONS has already had the scratchpad
    # copied over the running configuration at this point.
    with open(ETC_FRR_DAEMONS) as f:
        if re.search(r"""^\s*\S+\s*=\s*yes\b""", f.read()) is None:
            # No services are running.
            return "STOP"

    # Else, some services are still running, and none are being started.  Do nothing.
    return None


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format="%(asctime)s %(levelname)7s: %(message)s")
    doctest.testmod()
