"""
Copyright (C) 2017, 2018 Cumulus Networks, inc.

All Rights reserved.

This software is subject to the Cumulus Networks End User License Agreement available
at the following locations:.

Internet: https://cumulusnetworks.com/downloads/eula/latest/view/

Cumulus Linux systems: /usr/share/cumulus/EULA.txt

Settings managed in this module are set in accordance with this Mellanox documentation:
https://community.mellanox.com/docs/DOC-3036


Usage:
    net (add|del) forwarding ecmp hash-seed <0-4294967295>

Options:
    forwarding              : Packet forwarding parameters
    ecmp                    : Equal Cost MultiPath
    hash-seed               : Hash seed for ECMP entries
"""

from collections import OrderedDict
from nclu import (
    ConfigWrapperBase,
    ConfigFileKeyValue,
    globify_interface_commands,
    platform_vendor,
)
from nclu.wrappers.ifupdown2 import get_ifnames
from network_docopt import sort_ifaces
import logging
import os
import shutil

ETC_TRAFFIC = "/etc/cumulus/datapath/traffic.conf"
ETC_DATAPATH = "/usr/lib/python2.7/dist-packages/cumulus/__chip_config/mlx/datapath.conf"

log = logging.getLogger("netd")


def build_doc_string():
    global __doc__

    # storage-optimized is only supported on MLNX platforms.
    if platform_vendor() == "mlnx":
        mlnx_doc = """
Usage:
    net add interface <interface> storage-optimized [pfc]
    net del interface <interface> storage-optimized

Options:
    pfc                          : Priority Flow Control
    storage-optimized            : Enable RoCE
"""
        __doc__ += mlnx_doc

    return __doc__


def convert_list_string_to_list(list_string, list_is_ifnames):
    """
    traffic.conf has several keywords with a list as the value but is inconsistent
    with the format it uses to store the list.  Some of them use []s while others
    do not. Parse the various formats and return a list.
    """

    if list_string is None:
        return []

    assert isinstance(list_is_ifnames, bool)
    assert isinstance(list_string, basestring)

    # Standardize the format of list_string by:
    # - remove leading and trailing whitespaces
    # - remove leading and trailing []s
    # - replace commas with whitespaces
    # - split by whitespace and then rejoin with a single whitespace
    list_string = ' '.join((list_string.strip(" []").replace(',', ' ').split()))
    result = list_string.split()

    # Remove duplicate values within the list.
    result = list(set(result))

    # If these are interface names, sort accordingly.
    return sort_ifaces(result) if list_is_ifnames else sorted(result)


class ConfigWrapper(ConfigWrapperBase):

    def __init__(self, show_linux_command, color_diffs):
        super(ConfigWrapper, self).__init__("traffic", show_linux_command, color_diffs)
        self.files[ETC_TRAFFIC] =\
            ConfigFileKeyValue(self, ETC_TRAFFIC,
                       [["/bin/systemctl", "enable", "switchd.service"],
                        ["/bin/systemctl", "reset-failed", "switchd.service"],
                        ["/bin/systemctl", "restart", "switchd.service"]])

        self.files[ETC_DATAPATH] =\
            ConfigFileKeyValue(self, ETC_DATAPATH,
                       [["/bin/systemctl", "enable", "switchd.service"],
                        ["/bin/systemctl", "reset-failed", "switchd.service"],
                        ["/bin/systemctl", "restart", "switchd.service"]])

    def get_conf_file_key_value(self, filename, key):
        cfg = self.files[filename]
        cfg_content = cfg.get_value()
        return cfg_content.get(key)

    def get_storage_optimzed_ifnames(self):
        """
        Return a 2-tuple of two lists.  The first list includes interface names that have been configured for storage-optimized.
        The second list includes associated ecn port settings.
        """

        cfg_content = self.files[ETC_TRAFFIC].get_value()

        pfc_port_set = cfg_content.get("pfc.pfc_port_group.port_set")
        pfc_port_set = convert_list_string_to_list(pfc_port_set, True)

        ecn_port_set = cfg_content.get("ecn_red.port_set")
        ecn_port_set = convert_list_string_to_list(ecn_port_set, True)

        return (pfc_port_set, ecn_port_set)

    def show_config_commands(self, user_may_edit, ifupdown2_wrapper):
        supported = []
        interface_commands = []
        unsupported = []
        (pfc_port_set, ecn_port_set) = self.get_storage_optimzed_ifnames()

        for ifname in pfc_port_set:
            interface_commands.append("net add interface {0} storage-optimized pfc".format(ifname))

        for ifname in ecn_port_set:
            if ifname not in pfc_port_set:
                interface_commands.append("net add interface {0} storage-optimized".format(ifname))

        interface_commands = globify_interface_commands(interface_commands)
        supported.extend(interface_commands)
        ecmp_hash_seed = self.get_conf_file_key_value(ETC_TRAFFIC, "ecmp_hash_seed")

        if ecmp_hash_seed is not None:
            supported.append("net add forwarding ecmp hash-seed {0}".format(ecmp_hash_seed))

        return (supported, unsupported)

    def show_config_summary(self, user_may_edit, summary):
        (pfc_port_set, ecn_port_set) = self.get_storage_optimzed_ifnames()

        for ifname in pfc_port_set + ecn_port_set:
            ifname = "INTERFACE:{0}".format(ifname)

            if ifname not in summary:
                summary[ifname] = OrderedDict()

            if "COMMAND_LIST" not in summary[ifname]:
                summary[ifname]["COMMAND_LIST"] = []

        for ifname in pfc_port_set:
            ifname = "INTERFACE:{0}".format(ifname)
            summary[ifname]["COMMAND_LIST"].append("storage-optimized pfc")

        for ifname in ecn_port_set:
            if ifname not in pfc_port_set:
                ifname = "INTERFACE:{0}".format(ifname)
                summary[ifname]["COMMAND_LIST"].append("storage-optimized")

        ecmp_hash_seed = self.get_conf_file_key_value(ETC_TRAFFIC, "ecmp_hash_seed")

        if ecmp_hash_seed is not None:
            if "forwarding" not in summary:
                summary["forwarding"] = []

            summary["forwarding"].append("ecmp hash-seed {0}".format(ecmp_hash_seed))

    def get_pending(self):
        """
        Return the pending diff as a string.  If there are no changes, return the empty string.
        """

        result = ConfigWrapperBase.get_pending(self)

        if result:
            result += """\n\033[91m\n
********************************************************************
********************************************************************
NOTE: "net commit" will run "systemctl restart switchd" to apply the
changes to {0}. This will be service impacting.
********************************************************************
********************************************************************
\033[0m
""".format(ETC_TRAFFIC)

        return result

    def del_all(self, cli):
        """ Remove all RoCE related configuration. """

        for etc_filename, cfg in self.files.iteritems():

            if not os.path.isfile(cfg.scratchpad):
                shutil.copyfile(cfg.filename, cfg.scratchpad)

            cfg_content = cfg.get_value(True)

            if etc_filename == ETC_TRAFFIC:
                # Remove RoCE configuration from traffic.conf.
                self.del_storage_optimized_traffic_conf(cfg, cfg_content)

            elif etc_filename == ETC_DATAPATH:
                # Remove RoCE configuration from datapath.conf.
                self.del_storage_optimized_datapath_conf(cfg, cfg_content)

    def del_pfc_storage_optimized_traffic_conf(self, cfg, cfg_content):
        """ priority flow control """

        # Remove ROCE_PFC from pfc.port_group_list
        pfc_port_group_list = cfg_content.get("pfc.port_group_list")
        pfc_port_group_list = convert_list_string_to_list(pfc_port_group_list, False)

        if "ROCE_PFC" in pfc_port_group_list:
            pfc_port_group_list.remove("ROCE_PFC")

        if pfc_port_group_list:
            cfg.edit("pfc.port_group_list", "[{0}]".format(','.join(pfc_port_group_list)), give_feedback=False)
        else:
            cfg.delete("pfc.port_group_list")

        # Delete all ROCE_PFC configuration.
        for key in cfg_content:
            if "ROCE_PFC" in key:
                cfg.delete(key)

    def del_ecn_storage_optimized_traffic_conf(self, cfg, cfg_content):
        """ explicit congestion notification """

        # Remove ROCE_ECN from ecn_red.port_group_list
        ecn_red_port_group_list = cfg_content.get("ecn_red.ecn_red_port_group.port_group_list")
        ecn_red_port_group_list = convert_list_string_to_list(ecn_red_port_group_list, False)

        if "ROCE_ECN" in ecn_red_port_group_list:
            ecn_red_port_group_list.remove("ROCE_ECN")

        if ecn_red_port_group_list:
            cfg.edit("ecn_red.ecn_red_port_group.port_group_list", "[{0}]".format(','.join(ecn_red_port_group_list)), give_feedback=False)
        else:
            cfg.delete("ecn_red.ecn_red_port_group.port_group_list")

        # Delete all ROCE_ECN configuration.
        for key in cfg_content:
            if "ROCE_ECN" in key:
                cfg.delete(key)

    def del_storage_optimized_traffic_conf(self, cfg, cfg_content):
        self.del_pfc_storage_optimized_traffic_conf(cfg, cfg_content)
        self.del_ecn_storage_optimized_traffic_conf(cfg, cfg_content)
        cfg.delete("ecmp_hash_seed")

    def del_storage_optimized_datapath_conf(self, cfg, cfg_content):
        """
        If traffic.conf has no remaining RoCE interfaces configured, put
        datapath.conf back to some sane default. The default values hardcoded here
        were the mlnx datapath.conf defaults as of CL 3.5. This is not a perfect
        solution, but it is better than leaving datapath.conf setup for RoCE.
        """

        cfg.delete("priority_group.control.id")
        cfg.delete("priority_group.service.id")
        cfg.delete("priority_group.bulk.id")

        cfg.delete("priority_group.control.service_pool")
        cfg.delete("priority_group.service.service_pool")
        cfg.delete("priority_group.bulk.service_pool")

        cfg.delete("ingress_service_pool.0.mode")
        cfg.delete("egress_service_pool.0.mode")

        cfg.delete("flow_control.service_pool")

        cfg.edit("ingress_service_pool.0.percent", "100.0", give_feedback=False)
        cfg.delete("ingress_service_pool.1.percent")
        cfg.delete("ingress_service_pool.1.mode")

        cfg.edit("egress_service_pool.0.percent", "100.0", give_feedback=False)
        cfg.delete("egress_service_pool.1.percent")
        cfg.delete("egress_service_pool.1.mode")

        cfg.delete("priority_group.control.ingress_buffer.dynamic_quota")
        cfg.delete("priority_group.service.ingress_buffer.dynamic_quota")
        cfg.delete("priority_group.bulk.ingress_buffer.dynamic_quota")
        cfg.delete("flow_control.ingress_buffer.dynamic_quota")

        cfg.delete("priority_group.bulk.egress_buffer.uc.sp_dynamic_quota")
        cfg.delete("priority_group.service.egress_buffer.uc.sp_dynamic_quota")
        cfg.delete("priority_group.control.egress_buffer.uc.sp_dynamic_quota")

        cfg.delete("priority_group.bulk.egress_buffer.mc.sp_dynamic_quota")
        cfg.delete("priority_group.service.egress_buffer.mc.sp_dynamic_quota")
        cfg.delete("priority_group.control.egress_buffer.mc.sp_dynamic_quota")

    def eval_storage_optimized_traffic_conf(self, cli, ifnames):
        """
        Handle all add/del for the "storage-optimized" macro for the traffic.conf file.
        """

        tokens = cli.args
        cfg = self.files[ETC_TRAFFIC]
        cfg_content = cfg.get_value(True)
        command_ok = False
        roce_configured = True

        # These "port_set" variables are lists; not sets.

        pfc_port_set = cfg_content.get("pfc.pfc_port_group.port_set")
        pfc_port_set = convert_list_string_to_list(pfc_port_set, True)

        ecn_port_set = cfg_content.get("ecn_red.ecn_red_port_group.port_set")
        ecn_port_set = convert_list_string_to_list(ecn_port_set, True)

        if "add" in tokens:
            command_ok = True

            # =====================================
            # 1. Enable ECN and RED for priority 3.
            # =====================================

            # Add ROCE_ECN to ecn_red.port_group_list if it isn't already there.
            ecn_red_port_group_list = cfg_content.get("ecn_red.port_group_list")
            ecn_red_port_group_list = convert_list_string_to_list(ecn_red_port_group_list, False)

            if "ROCE_ECN" not in ecn_red_port_group_list:
                ecn_red_port_group_list.append("ROCE_ECN")

            cfg.edit("ecn_red.port_group_list", "[{0}]".format(','.join(ecn_red_port_group_list)), give_feedback=False)

            # Set cos_list.
            cfg.edit("ecn_red.ecn_red_port_group.cos_list", "[3]", give_feedback=False)

            # Add each ifname to ecn_red.port_set, which is a list that might not be empty.
            for ifname in ifnames:
                if ifname not in ecn_port_set:
                    ecn_port_set.append(ifname)

            ecn_port_set = sort_ifaces(ecn_port_set)
            cfg.edit("ecn_red.ecn_red_port_group.port_set", ','.join(ecn_port_set), give_feedback=False)

            # Misc ECN RED settings
            cfg.edit("ecn_red.ecn_red_port_group.ecn_enable", "true", give_feedback=False)
            cfg.edit("ecn_red.ecn_red_port_group.red_enable", "false", give_feedback=False)
            cfg.edit("ecn_red.ecn_red_port_group.min_threshold_bytes", "153600", give_feedback=False)
            cfg.edit("ecn_red.ecn_red_port_group.max_threshold_bytes", "1536000", give_feedback=False)
            cfg.edit("ecn_red.ecn_red_port_group.probability", "100", give_feedback=False)

            # ======================================================
            # 2. Set trust mode to DSCP, and map DSCP values to COS.
            # ======================================================

            cfg.edit("traffic.packet_priority_source_set", "[dscp]", give_feedback=False)

            priority_source = range(26) + range(27, 48) + range(49, 64)
            cfg.edit("traffic.cos_0.priority_source.dscp", str(priority_source), give_feedback=False)

            cfg.edit("traffic.cos_1.priority_source.dscp", "[]", give_feedback=False)
            cfg.edit("traffic.cos_2.priority_source.dscp", "[48]", give_feedback=False)  # For CNPs
            cfg.edit("traffic.cos_3.priority_source.dscp", "[26]", give_feedback=False)  # For RoCE
            cfg.edit("traffic.cos_4.priority_source.dscp", "[]", give_feedback=False)
            cfg.edit("traffic.cos_5.priority_source.dscp", "[]", give_feedback=False)
            cfg.edit("traffic.cos_6.priority_source.dscp", "[]", give_feedback=False)
            cfg.edit("traffic.cos_7.priority_source.dscp", "[]", give_feedback=False)

            # ============================================
            # 3. Map switch priorities to priority groups.
            # ============================================

            cfg.edit("traffic.priority_group_list", "[control, service, bulk]", give_feedback=False)
            cfg.edit("priority_group.control.cos_list", "[2]", give_feedback=False)
            cfg.edit("priority_group.service.cos_list", "[3]", give_feedback=False)
            cfg.edit("priority_group.bulk.cos_list", "[0,1,4,5,6,7]", give_feedback=False)

            if "pfc" in tokens:
                # =============================
                # 4. Enable PFC for priority 3.
                # =============================

                # Add ROCE_PFC to pfc.port_group_list (if it isn't already there)
                pfc_port_group_list = cfg_content.get("pfc.port_group_list")
                pfc_port_group_list = convert_list_string_to_list(pfc_port_group_list, False)

                if "ROCE_PFC" not in pfc_port_group_list:
                    pfc_port_group_list.append("ROCE_PFC")

                cfg.edit("pfc.port_group_list", "[{0}]".format(','.join(pfc_port_group_list)), give_feedback=False)

                # set the cos_list
                cfg.edit("pfc.pfc_port_group.cos_list", "[3]", give_feedback=False)

                # Add each ifname to pfc.pfc_port_group.port_set, which is a list that might not be empty.
                for ifname in ifnames:
                    if ifname not in pfc_port_set:
                        pfc_port_set.append(ifname)

                pfc_port_set = sort_ifaces(pfc_port_set)
                command_ok = cfg.edit("pfc.pfc_port_group.port_set", ','.join(pfc_port_set))

                # Misc PFC settings
                cfg.edit("pfc.pfc_port_group.port_buffer_bytes", "70000", give_feedback=False)
                cfg.edit("pfc.pfc_port_group.xoff_size", "18000", give_feedback=False)
                cfg.edit("pfc.pfc_port_group.xon_delta", "0", give_feedback=False)
                cfg.edit("pfc.pfc_port_group.tx_enable", "true", give_feedback=False)
                cfg.edit("pfc.pfc_port_group.rx_enable", "true", give_feedback=False)

            # =============================================================================
            # 5. Configure weighted round-robin to set the traffic group scheduling weight.
            # =============================================================================

            cfg.edit("scheduling.algorithm", "dwrr", give_feedback=False)
            cfg.edit("priority_group.control.weight", "0", give_feedback=False)
            cfg.edit("priority_group.service.weight", "16", give_feedback=False)
            cfg.edit("priority_group.bulk.weight", "16", give_feedback=False)

        elif "del" in tokens:

            for ifname in ifnames:
                if ifname in pfc_port_set:
                    pfc_port_set.remove(ifname)
                    command_ok = True

                if ifname in ecn_port_set:
                    ecn_port_set.remove(ifname)
                    command_ok = True

            if not command_ok:
                print "{0} does not have storage-optimized configured for {1}.".format(ETC_TRAFFIC, ", ".join(ifnames))

            if pfc_port_set or ecn_port_set:

                if pfc_port_set:
                    cfg.edit("pfc.pfc_port_group.port_set", ','.join(pfc_port_set), give_feedback=False)
                else:
                    self.del_pfc_storage_optimized_traffic_conf(cfg, cfg_content)

                if ecn_port_set:
                    cfg.edit("ecn_red.ecn_red_port_group.port_set", ','.join(ecn_port_set), give_feedback=False)
                else:
                    self.del_ecn_storage_optimized_traffic_conf(cfg, cfg_content)

            else:
                # There are no interfaces configured for ROCE.  Delete the RoCE related configuration.
                self.del_storage_optimized_traffic_conf(cfg, cfg_content)
                roce_configured = False

        return (command_ok, roce_configured)

    def eval_storage_optimized_datapath_conf(self, cli, roce_configured):
        """
        Handle all add/del for the "storage-optimized" macro for the datapath.conf file.
        """

        tokens = cli.args
        cfg = self.files[ETC_DATAPATH]
        cfg_content = cfg.get_value(True)

        if "add" in tokens:

            pfc_port_set, _ = self.get_storage_optimzed_ifnames()

            # =============================================
            # 6. Assign group IDs, and create buffer pools.
            # =============================================

            cfg.edit("priority_group.control.id", '0', give_feedback=False)
            cfg.edit("priority_group.service.id", '0', give_feedback=False)
            cfg.edit("priority_group.bulk.id", '0', give_feedback=False)

            cfg.edit("priority_group.control.service_pool", '0', give_feedback=False)
            cfg.edit("priority_group.service.service_pool", '0', give_feedback=False)
            cfg.edit("priority_group.bulk.service_pool", '0', give_feedback=False)

            cfg.edit("ingress_service_pool.0.mode", '1', give_feedback=False)
            cfg.edit("egress_service_pool.0.mode", '1', give_feedback=False)

            if "pfc" in tokens or pfc_port_set:
                cfg.edit("flow_control.service_pool", '1', give_feedback=False)

                cfg.edit("ingress_service_pool.0.percent", "50.0", give_feedback=False)
                cfg.edit("ingress_service_pool.1.percent", "50.0", give_feedback=False)
                cfg.edit("ingress_service_pool.1.mode", '1', give_feedback=False)

                cfg.edit("egress_service_pool.0.percent", "50.0", give_feedback=False)
                cfg.edit("egress_service_pool.1.percent", "100.0", give_feedback=False)
                cfg.edit("egress_service_pool.1.mode", '1', give_feedback=False)
            else:
                cfg.delete("flow_control.service_pool")

                cfg.edit("ingress_service_pool.0.percent", "100.0", give_feedback=False)
                cfg.delete("ingress_service_pool.1.percent")
                cfg.delete("ingress_service_pool.1.mode")

                cfg.edit("egress_service_pool.0.percent", "100.0", give_feedback=False)
                cfg.delete("egress_service_pool.1.percent")
                cfg.delete("egress_service_pool.1.mode")

            # ==========================
            # 7. Configure alpha values.
            # ==========================

            # Ingress buffer
            cfg.edit("priority_group.control.ingress_buffer.dynamic_quota", "11", give_feedback=False)
            cfg.edit("priority_group.service.ingress_buffer.dynamic_quota", "11", give_feedback=False)
            cfg.edit("priority_group.bulk.ingress_buffer.dynamic_quota", "11", give_feedback=False)
            if "pfc" in tokens or pfc_port_set:
                cfg.edit("flow_control.ingress_buffer.dynamic_quota", '9', give_feedback=False)  # RoCE

            # Unicast egress buffer
            cfg.edit("priority_group.bulk.egress_buffer.uc.sp_dynamic_quota", "11", give_feedback=False)
            cfg.edit("priority_group.service.egress_buffer.uc.sp_dynamic_quota", "11", give_feedback=False)
            cfg.edit("priority_group.control.egress_buffer.uc.sp_dynamic_quota", "11", give_feedback=False)

            # Multicast egress buffer
            cfg.edit("priority_group.bulk.egress_buffer.mc.sp_dynamic_quota", '9', give_feedback=False)
            cfg.edit("priority_group.service.egress_buffer.mc.sp_dynamic_quota", '9', give_feedback=False)
            cfg.edit("priority_group.control.egress_buffer.mc.sp_dynamic_quota", '9', give_feedback=False)

        elif "del" in tokens:
            if not roce_configured:
                self.del_storage_optimized_datapath_conf(cfg, cfg_content)

        return True

    def eval_storage_optimized(self, cli):
        """
        Handle all add/del for the "storage-optimized" macro.
        """

        ifnames_valid, ifnames = get_ifnames(cli)

        if not ifnames_valid:
            return False

        command_ok, roce_configured = self.eval_storage_optimized_traffic_conf(cli, ifnames)

        if command_ok:
            command_ok = self.eval_storage_optimized_datapath_conf(cli, roce_configured)

        return command_ok

    def eval_ecmp_hash_seed(self, cli):
        """
        net (add|del) forwarding ecmp hash-seed <0-4294967295>

        Handle all add/del for the "ecmp hash-seed" macro.
        """

        # Convenience variable
        tokens = cli.args
        cfg = self.files[ETC_TRAFFIC]

        if "add" in tokens:
            assert "<0-4294967295>" in tokens, "command: {0}".format(cli.argv_expanded)
            return cfg.edit("ecmp_hash_seed", tokens["<0-4294967295>"])
        elif "del" in tokens:
            return cfg.delete("ecmp_hash_seed")

        log.debug("NCLU does not know how to process command \"{0}\".".format(cli.argv_expanded))
        return False

    def eval_command_line_args(self, cli):
        super(ConfigWrapper, self).eval_command_line_args(cli)

        # Convenience variable
        tokens = cli.args

        if "storage-optimized" in tokens:
            return self.eval_storage_optimized(cli)
        elif "forwarding" in tokens and "ecmp" in tokens and "hash-seed" in tokens:
            return self.eval_ecmp_hash_seed(cli)

        log.debug("NCLU does not know how to process command \"{0}\".".format(cli.argv_expanded))
        return False
